package jp.sourceforge.armadillo;

import java.io.*;
import java.nio.channels.*;

import jp.sourceforge.armadillo.io.*;
import jp.sourceforge.armadillo.lzh.*;

/**
 * LZHA[JCoB
 */
public final class LzhArchiver implements Archiver {

    private static final byte HEADER_LEVEL = 1;

    private LzhOutputStream los;
    private LzhHeader header;
    private RandomAccessFile rfile;

    /**
     * LzhArchiver̐B
     * @param os OutputStream
     */
    public LzhArchiver(OutputStream os) {
        this.los = new LzhOutputStream(os);
    }

    /**
     * LzhArchiver̐B
     * @param file o͐t@C
     * @throws IOException I/OG[
     */
    public LzhArchiver(File file) throws IOException {
        RandomAccessFile r = new RandomAccessFile(file, "rw");
        try {
            r.setLength(0L); // overwrite
            this.los = new LzhOutputStream(Channels.newOutputStream(r.getChannel()));
            this.header = new LzhHeader();
            this.rfile = r;
        } catch (Exception ex) {
            r.close();
            IOException e = new IOException();
            e.initCause(ex);
            throw e;
        }
    }

    /* @see jp.sourceforge.armadillo.Archiver#addEntry(jp.sourceforge.armadillo.ArchiveEntry, java.io.File) */
    public void addEntry(ArchiveEntry ae, File file) throws IOException {
        if (file.isDirectory()) {
            addDirectoryEntry(ae);
        } else {
            addFileEntry(ae, file);
        }
        ae.setAdded(true);
    }

    /* @see jp.sourceforge.armadillo.Archiver#addEntry(jp.sourceforge.armadillo.ArchiveEntry,
                                                       java.io.InputStream, long) */
    public void addEntry(ArchiveEntry ae, InputStream is, long length) throws IOException {
        if (ae.isDirectory()) {
            addDirectoryEntry(ae);
        } else {
            File file = IOUtilities.createTemporaryFile();
            if (length > 0 && is != null) {
                OutputStream os = new FileOutputStream(file);
                try {
                    IOUtilities.transfer(is, os, length);
                } finally {
                    os.close();
                }
            }
            addFileEntry(ae, file);
            file.delete();
        }
        ae.setAdded(true);
    }

    /**
     * fBNgGg̒ǉB
     * @param ae ArchiveEntry
     * @throws IOException o̓G[ꍇ
     */
    private void addDirectoryEntry(ArchiveEntry ae) throws IOException {
        LzhEntry entry = new LzhEntry(ae.getName());
        entry.setHeaderLevel(HEADER_LEVEL);
        entry.setMethod(LzhMethod.LHD);
        entry.setCrc((short)0);
        entry.setSize(0L);
        entry.setCompressedSize(0L);
        entry.setLastModified(ae.getLastModified());
        los.putNextEntry(entry);
        los.closeEntry();
        ae.setCrc(0L);
        ae.setSize(0L);
        ae.setCompressedSize(0L);
    }

    /**
     * t@CGg̒ǉB
     * @param ae ArchiveEntry
     * @param file t@C
     * @throws IOException o̓G[ꍇ
     */
    private void addFileEntry(ArchiveEntry ae, File file) throws IOException {
        LzhEntry entry = new LzhEntry(ae.getName());
        entry.setHeaderLevel(HEADER_LEVEL);
        entry.setOsIdentifier('J');
        entry.setLastModified(ae.getLastModified());
        final long fileSize = file.length();
        if (fileSize > 0) {
            entry.setMethod(LzhMethod.LH5);
            entry.setSize(fileSize);
            RandomAccessFile r = new RandomAccessFile(file, "r");
            try {
                long size;
                InputStream is = Channels.newInputStream(r.getChannel());
                if (rfile == null) {
                    try {
                        tryOut(entry, is);
                        if (entry.getCompressedSize() >= fileSize) {
                            entry.setMethod(LzhMethod.LH0);
                        }
                    } catch (LzhQuit ex) {
                        entry.setMethod(LzhMethod.LH0);
                    }
                    r.seek(0L);
                    los.putNextEntry(entry);
                    size = IOUtilities.transfer(is, los, fileSize);
                    los.closeEntry();
                } else {
                    los.flush();
                    long first = rfile.getFilePointer();
                    los.putNextEntry(entry);
                    long eoh = rfile.getFilePointer();
                    boolean quit;
                    try {
                        size = IOUtilities.transfer(is, los, fileSize);
                        quit = false;
                    } catch (LzhQuit ex) {
                        size = 0;
                        quit = true;
                    }
                    los.closeEntry();
                    entry.setCrc(los.getCrc());
                    long last = rfile.getFilePointer();
                    long compressedSize = last - eoh;
                    if (compressedSize >= fileSize || quit) {
                        entry.setMethod(LzhMethod.LH0);
                        entry.setCompressedSize(fileSize);
                        r.seek(0L);
                        rfile.seek(first);
                        los.putNextEntry(entry);
                        size = IOUtilities.transfer(is, los, fileSize);
                        long p = rfile.getFilePointer();
                        entry.setCrc(los.getCrc());
                        rfile.seek(first);
                        los.putNextEntry(entry);
                        rfile.seek(p);
                        los.closeEntry();
                    } else {
                        entry.setCompressedSize(compressedSize);
                        rfile.seek(first);
                        header.write(Channels.newOutputStream(rfile.getChannel()), entry);
                        rfile.seek(last);
                    }
                }
                assert size == fileSize;
                ae.setCrc(entry.getCrc());
                ae.setSize(entry.getSize());
                ae.setCompressedSize(entry.getCompressedSize());
            } finally {
                r.close();
            }
        } else {
            los.putNextEntry(entry);
            los.closeEntry();
            ae.setCrc(0L);
            ae.setSize(0L);
            ae.setCompressedSize(0L);
        }
    }

    /**
     * ksB
     * @param entry LzhEntry
     * @param is InputStream
     * @throws IOException o̓G[ꍇ
     */
    private static void tryOut(LzhEntry entry, InputStream is) throws IOException {
        SizeDetectionOutputStream detector = new SizeDetectionOutputStream();
        LzhOutputStream los = new LzhOutputStream(detector);
        try {
            los.putNextEntry(entry);
            IOUtilities.transferAll(is, los);
            los.closeEntry();
        } finally {
            los.close();
        }
        entry.setCrc(los.getCrc());
        entry.setCompressedSize(detector.getSize());
        assert entry.getCompressedSize() == detector.getSize();
    }

    /* @see jp.sourceforge.armadillo.Archiver#close() */
    public void close() throws IOException {
        try {
            los.close();
        } finally {
            if (rfile != null) {
                rfile.close();
            }
        }
    }

}
